package dev.fitko.fitconnect.client.sender.strategies;

import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.EncryptedAttachment;
import dev.fitko.fitconnect.client.sender.model.SendableEncryptedSubmission;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.UUID;

import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildAuthenticationTags;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildAuthenticationTagsForEncryptedSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildCreateSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSentSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSubmitSubmission;

public class SendEncryptedSubmissionStrategy {

    private static final Logger LOGGER = LoggerFactory.getLogger(SendEncryptedSubmissionStrategy.class);

    private final Sender sender;

    public SendEncryptedSubmissionStrategy(final Sender sender) {
        this.sender = sender;
    }

    public SentSubmission send(final SendableEncryptedSubmission submissionPayload) {

        try {

            final List<EncryptedAttachment> encryptedAttachments = submissionPayload.getAttachments();
            final String encryptedData = submissionPayload.getData();
            final String encryptedMetadata = submissionPayload.getMetadata();

            final UUID submissionId = announceNewSubmission(submissionPayload, encryptedAttachments);
            final SubmitSubmission submitSubmission = buildSubmitSubmission(submissionId, encryptedData, encryptedMetadata);

            final var startTimeAttachmentUpload = StopWatch.start();
            uploadAttachments(encryptedAttachments, submissionId);
            LOGGER.info("Uploading attachments took {}", StopWatch.stopWithFormattedTime(startTimeAttachmentUpload));

            final var startTimeSubmissionUpload = StopWatch.start();
            final Submission submission = sender.sendSubmission(submitSubmission);
            LOGGER.info("Uploading submission took {}", StopWatch.stopWithFormattedTime(startTimeSubmissionUpload));

            LOGGER.info("SUCCESSFULLY HANDED IN SUBMISSION !");
            final AuthenticationTags authenticationTags = buildAuthenticationTagsForEncryptedSubmission(encryptedAttachments, encryptedData, encryptedMetadata);
            return buildSentSubmission(authenticationTags, submission);

        } catch (final RestApiException e) {
            LOGGER.error("Sending submission failed", e);
        } catch (final SubmissionNotCreatedException e) {
            LOGGER.error("Failed to announce new submission", e);
        }

        return null;
    }

    private UUID announceNewSubmission(final SendableEncryptedSubmission submissionPayload, final List<EncryptedAttachment> encryptedAttachments) {
        final CreateSubmission submissionToAnnounce = buildCreateSubmission(submissionPayload, encryptedAttachments);
        return sender.createSubmission(submissionToAnnounce).getSubmissionId();
    }

    private void uploadAttachments(final List<EncryptedAttachment> encryptedAttachments, final UUID submissionId) {
        if (encryptedAttachments.isEmpty()) {
            LOGGER.info("No attachments to upload");
        } else {
            LOGGER.info("Uploading {} attachment(s)", encryptedAttachments.size());
            encryptedAttachments.forEach(a -> sender.uploadAttachment(submissionId, a.getAttachmentId(), a.getContent()));
        }
    }
}
