package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.SendableEncryptedSubmission;
import dev.fitko.fitconnect.client.sender.model.SendableSubmission;

import java.net.URI;
import java.util.Objects;
import java.util.UUID;
import java.util.regex.Pattern;

public class ValidDataGuard {

    private static final Pattern LEIKA_KEY_PATTERN = Pattern.compile("^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$");
    private final Sender sender;

    public ValidDataGuard(final Sender sender) {
        this.sender = sender;
    }

    /**
     * Checks if the encrypted submission payload is in a valid state for sending.
     *
     * @param sendableEncryptedSubmission payload to be checked
     * @throws IllegalArgumentException if one of the checks fails
     * @throws IllegalStateException    if the destination does not have a declared service type
     */
    public void ensureValidDataPayload(final SendableEncryptedSubmission sendableEncryptedSubmission) {
        if (sendableEncryptedSubmission == null) {
            throw new IllegalArgumentException("Encrypted payload must not be null.");
        }
        if (sendableEncryptedSubmission.getData() == null) {
            throw new IllegalArgumentException("Encrypted data is mandatory, but was null.");
        }
        if (sendableEncryptedSubmission.getMetadata() == null) {
            throw new IllegalArgumentException("Encrypted metadata must not be null.");
        }
        testDefaults(sendableEncryptedSubmission.getDestinationId(), sendableEncryptedSubmission.getServiceName(), sendableEncryptedSubmission.getServiceIdentifier());
    }

    /**
     * Checks if the unencrypted submission payload is in a valid state for sending.
     *
     * @param sendableSubmission payload to be checked
     * @throws IllegalArgumentException if one of the checks fails
     */
    public void ensureValidDataPayload(final SendableSubmission sendableSubmission) {
        if (sendableSubmission == null) {
            throw new IllegalArgumentException("Payload must not be null.");
        }
        if (sendableSubmission.getData() == null) {
            throw new IllegalArgumentException("Data is mandatory, but was null.");
        }
        testOnValidDataFormat(sendableSubmission);
        testDefaults(sendableSubmission);
    }

    private void testDefaults(final UUID destinationId, final String serviceName, final String serviceIdentifier) {
        if (destinationId == null) {
            throw new IllegalArgumentException("DestinationId is mandatory, but was null.");
        } else if (serviceIdentifier == null) {
            throw new IllegalArgumentException("Leika key is mandatory, but was null.");
        } else if (noValidLeikaKeyPattern(serviceIdentifier)) {
            throw new IllegalArgumentException("LeikaKey has invalid format, please follow: ^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$.");
        }

        Destination destination = sender.getDestination(destinationId);
        if (serviceTypeDoesNotMatchDestination(destination, serviceIdentifier)) {
            throw new IllegalArgumentException("Provided service type '" + serviceIdentifier + "' is not allowed by the destination ");
        }
    }

    private void testDefaults(final SendableSubmission sendableSubmission) {

        if (sendableSubmission.getDestinationId() == null) {
            throw new IllegalArgumentException("DestinationId is mandatory, but was null.");
        } else if (sendableSubmission.getServiceIdentifier() == null) {
            throw new IllegalArgumentException("Leika key is mandatory, but was null.");
        } else if (noValidLeikaKeyPattern(sendableSubmission.getServiceIdentifier())) {
            throw new IllegalArgumentException("LeikaKey has invalid format, please follow: ^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$.");
        }

        Destination destination = sender.getDestination(sendableSubmission.getDestinationId());
        if (serviceTypeDoesNotMatchDestination(destination, sendableSubmission.getServiceIdentifier())) {
            throw new IllegalArgumentException("Provided service type '" + sendableSubmission.getServiceIdentifier() + "' is not allowed by the destination ");
        }
        final MimeType dataMimeType = MimeType.fromValue(sendableSubmission.getDataMimeType());
        if (mimeTypeAndSchemaUriDoNotMatchDestination(destination, dataMimeType, sendableSubmission.getSchemaUri())) {
            throw new IllegalArgumentException("Combination of provided MIME type '" + sendableSubmission.getDataMimeType() +
                    "' and schema URI '" + sendableSubmission.getSchemaUri() + "' is not allowed by the destination");
        }
    }

    private boolean serviceTypeDoesNotMatchDestination(final Destination destination, final String serviceIdentifier) {
        return destination.getServices().stream()
                .map(DestinationService::getIdentifier)
                .filter(Objects::nonNull)
                .filter(destinationServiceIdentifier -> destinationServiceIdentifier.equals(serviceIdentifier))
                .findFirst()
                .isEmpty();
    }

    private boolean mimeTypeAndSchemaUriDoNotMatchDestination(final Destination destination, final MimeType mimeType, final URI schemaUri) {
        return destination.getServices().stream()
                .flatMap(service -> service.getSubmissionSchemas().stream())
                .filter(Objects::nonNull)
                .filter(schema -> schema.getMimeType().equals(mimeType))
                .filter(schema -> schema.getSchemaUri().equals(schemaUri))
                .findFirst()
                .isEmpty();
    }

    private void testOnValidDataFormat(final SendableSubmission sendableSubmission) {
        final MimeType dataMimeType = MimeType.fromValue(sendableSubmission.getDataMimeType());
        if (dataMimeType.equals(MimeType.APPLICATION_JSON)) {
            checkJsonFormat(sendableSubmission);
        } else if (dataMimeType.equals(MimeType.APPLICATION_XML)) {
            checkXmlFormat(sendableSubmission);
        }
    }

    private void checkXmlFormat(final SendableSubmission sendableSubmission) {
        final ValidationResult validationResult = sender.validateXmlFormat(sendableSubmission.getData());
        if (validationResult.hasError()) {
            throw new IllegalArgumentException("Data is not in expected xml format, please provide valid xml: " + validationResult.getError().getMessage());
        }
    }

    private void checkJsonFormat(final SendableSubmission sendableSubmission) {
        final ValidationResult validationResult = sender.validateJsonFormat(sendableSubmission.getData());
        if (validationResult.hasError()) {
            throw new IllegalArgumentException("Data is not in expected json format, please provide valid json: " + validationResult.getError().getMessage());
        }
    }

    private boolean noValidLeikaKeyPattern(final String leikaKey) {
        if (leikaKey.length() < 7 || leikaKey.length() > 255) {
            return true;
        }
        return !LEIKA_KEY_PATTERN.matcher(leikaKey).matches();
    }
}
